
/*
 * Copyright (C) 2024 Web Server LLC
 */

#include <ngx_config.h>
#include <ngx_core.h>
#include <ngx_wasm.h>

#include <wasm.h>
#include <wasmtime.h>

#define ngx_wasmtime_trap(msg)  wasmtime_trap_new(msg, ngx_strlen(msg))

#define ngx_wasmtime_call_data(cl)                                            \
    wasmtime_context_get_data(                                                \
            (wasmtime_context_t *) wasmtime_caller_context(cl))


/* Host API modules require access to this to define own functions */
typedef struct {
    wasm_config_t           *config;
    wasm_engine_t           *engine;
    wasmtime_linker_t       *linker;
    ngx_uint_t               timedout;
    size_t                   stack_size;
    ngx_flag_t               enable_wasi;
} ngx_wasmtime_conf_t;


typedef struct {
    wasmtime_instance_t      instance;

    wasmtime_store_t        *store;
    wasmtime_context_t      *context;
    wasi_config_t           *wasi_config;

    wasmtime_memory_t        memory;
} ngx_wasmtime_instance_t;


typedef wasm_trap_t *(*ngx_wasmtime_host_handler_pt)(void *env,
    wasmtime_caller_t *caller, const wasmtime_val_t *args, size_t nargs,
    wasmtime_val_t *rc, size_t nres);


/* exported host function description */
typedef struct {
    ngx_str_t                          export;
    ngx_wasmtime_host_handler_pt       handler;
    ngx_uint_t                         nargs;
} ngx_wasmtime_host_handler_t;


static void ngx_wasmtime_log_error(ngx_uint_t level, ngx_log_t *log,
    wasmtime_error_t *err, wasm_trap_t *trap, char *fmt, ...);
static ngx_int_t ngx_wasmtime_lookup_func(ngx_wasm_instance_t *wi,
    ngx_str_t *name, wasmtime_func_t *func);

static void  *ngx_wasmtime_create_conf(ngx_cycle_t *cycle);
static char *ngx_wasmtime_init_conf(ngx_cycle_t *cycle, void *conf);
static ngx_int_t ngx_wasmtime_create_runtime(ngx_cycle_t *cycle);
static void ngx_wasmtime_delete_runtime(ngx_cycle_t *cycle);

static ngx_int_t ngx_wasmtime_create_module(ngx_cycle_t *cycle,
    ngx_wasm_module_t *wm);
static void ngx_wasmtime_delete_module(ngx_wasm_module_t *wm);

static ngx_wasm_instance_t *ngx_wasmtime_create_instance(
    ngx_wasm_module_t *wm, ngx_pool_t *pool, ngx_log_t *log);

static void ngx_wasmtime_delete_instance(ngx_wasm_instance_t *wi);

static ngx_int_t ngx_wasmtime_enter(ngx_wasm_instance_t *wi,
    ngx_wasm_call_t *call);

void *ngx_wasmtime_translate(ngx_wasm_instance_t *wi, uint32_t addr,
    size_t len);

static ngx_int_t ngx_wasmtime_init_host_interface(ngx_wasmtime_conf_t *wtcf,
    ngx_str_t *ns, ngx_wasmtime_host_handler_t *item, ngx_log_t *log);

static wasm_trap_t *ngx_wasmtime_open(void *env, wasmtime_caller_t *caller,
    const wasmtime_val_t *args, size_t nargs, wasmtime_val_t *rc, size_t nres);
static wasm_trap_t *ngx_wasmtime_close(void *env, wasmtime_caller_t *caller,
    const wasmtime_val_t *args, size_t nargs, wasmtime_val_t *rc, size_t nres);
static wasm_trap_t *ngx_wasmtime_read(void *env, wasmtime_caller_t *caller,
    const wasmtime_val_t *args, size_t nargs, wasmtime_val_t *rc, size_t nres);
static wasm_trap_t *ngx_wasmtime_write(void *env, wasmtime_caller_t *caller,
    const wasmtime_val_t *args, size_t nargs, wasmtime_val_t *rc, size_t nres);
static wasm_trap_t *ngx_wasmtime_get(void *env, wasmtime_caller_t *caller,
    const wasmtime_val_t *args, size_t nargs, wasmtime_val_t *rc, size_t nres);
static wasm_trap_t *ngx_wasmtime_set(void *env, wasmtime_caller_t *caller,
    const wasmtime_val_t *args, size_t nargs, wasmtime_val_t *rc, size_t nres);


static ngx_command_t  ngx_wasmtime_commands[] = {

    { ngx_string("wasmtime_stack_size"),
      NGX_WASM_CONF|NGX_CONF_TAKE1,
      ngx_conf_set_size_slot,
      0,
      offsetof(ngx_wasmtime_conf_t, stack_size),
      NULL },

    { ngx_string("wasmtime_enable_wasi"),
      NGX_WASM_CONF|NGX_CONF_FLAG,
      ngx_conf_set_flag_slot,
      0,
      offsetof(ngx_wasmtime_conf_t, enable_wasi),
      NULL },

      ngx_null_command
};

static ngx_str_t  wasmtime_name = ngx_string("wasmtime");

static ngx_wasm_core_module_t  ngx_wasmtime_module_ctx = {
    &wasmtime_name,
    ngx_wasmtime_create_conf,              /* create configuration */
    ngx_wasmtime_init_conf,                /* init configuration */

    {
        ngx_wasmtime_create_runtime,
        ngx_wasmtime_delete_runtime,
        ngx_wasmtime_create_module,
        ngx_wasmtime_delete_module,
        ngx_wasmtime_create_instance,
        ngx_wasmtime_delete_instance,
        ngx_wasmtime_enter,
        ngx_wasmtime_translate,
        /*
         * wasmtime starts threads - this leads to problems after fork(),
         * thus on-start module validation is disabled
         */
        0
    }
};


ngx_module_t  ngx_wasmtime_module = {
    NGX_MODULE_V1,
    &ngx_wasmtime_module_ctx,              /* module context */
    ngx_wasmtime_commands,                 /* module directives */
    NGX_WASM_MODULE,                       /* module type */
    NULL,                                  /* init master */
    NULL,                                  /* init module */
    NULL,                                  /* init process */
    NULL,                                  /* init thread */
    NULL,                                  /* exit thread */
    NULL,                                  /* exit process */
    NULL,                                  /* exit master */
    NGX_MODULE_V1_PADDING
};


static ngx_str_t ngx_wasi_prefix = ngx_string("ngx:wasi/syscall");

static ngx_wasmtime_host_handler_t ngx_wasmtime_host_handlers[] = {
    { ngx_string("open"), ngx_wasmtime_open, 3 },
    { ngx_string("close"), ngx_wasmtime_close, 1 },
    { ngx_string("read"), ngx_wasmtime_read, 3 },
    { ngx_string("write"), ngx_wasmtime_write, 3 },
    { ngx_string("get"), ngx_wasmtime_get, 4 },
    { ngx_string("set"), ngx_wasmtime_set, 4 },
};


static void
ngx_wasmtime_log_error(ngx_uint_t level, ngx_log_t *log, wasmtime_error_t *err,
    wasm_trap_t *trap, char *fmt, ...)
{
    va_list           args;
    u_char           *p, *last;
    u_char            errstr[NGX_MAX_CONF_ERRSTR];
    wasm_byte_vec_t   msg;

    last = errstr + NGX_MAX_CONF_ERRSTR;

    va_start(args, fmt);
    p = ngx_vslprintf(errstr, last, fmt, args);
    va_end(args);

    if (err != NULL) {
        wasmtime_error_message(err, &msg);

    } else {
        wasm_trap_message(trap, &msg);
    }

    if (msg.size == 0) {
        goto done;
    }

    p = ngx_slprintf(p, last, " (wasmtime %s: ",
                     err != NULL ? "error" : "trap");

    if (p + msg.size > last) {
        msg.size = last - p;
    }

    p = ngx_cpymem(p, (u_char *) msg.data, msg.size);

    if (p < last) {
        *p++ = ')';
    }

done:

    ngx_log_error(level, log, 0, "%*s", p - errstr, errstr);

    if (err != NULL) {
        wasmtime_error_delete(err);

    } else {
        wasm_trap_delete(trap);
    }
}


static void *
ngx_wasmtime_create_conf(ngx_cycle_t *cycle)
{
    ngx_wasmtime_conf_t  *wtcf;

    wtcf = ngx_pcalloc(cycle->pool, sizeof(ngx_wasmtime_conf_t));
    if (wtcf == NULL) {
        return NULL;
    }

    wtcf->stack_size = NGX_CONF_UNSET_SIZE;
    wtcf->enable_wasi = NGX_CONF_UNSET;

    return wtcf;
}


static char *
ngx_wasmtime_init_conf(ngx_cycle_t *cycle, void *conf)
{
    ngx_wasmtime_conf_t  *wtcf = conf;

    ngx_conf_init_size_value(wtcf->stack_size, 8092);
    ngx_conf_init_value(wtcf->enable_wasi, 1);

    return NGX_CONF_OK;
}


static ngx_int_t
ngx_wasmtime_create_runtime(ngx_cycle_t *cycle)
{
    ngx_uint_t            i, nhandlers;
    wasmtime_error_t     *error;
    ngx_wasmtime_conf_t  *wtcf;

    wtcf = (ngx_wasmtime_conf_t *) ngx_wasm_get_conf(cycle->conf_ctx,
                                                     ngx_wasmtime_module);

    wtcf->config = wasm_config_new();
    if (wtcf->config == NULL) {
        ngx_log_error(NGX_LOG_EMERG, cycle->log, 0,
                      "wasmtime: wasm_config_new() failed");
        return NGX_ERROR;
    }

    wasmtime_config_max_wasm_stack_set(wtcf->config, wtcf->stack_size);

    wtcf->engine = wasm_engine_new_with_config(wtcf->config);
    if (wtcf->engine == NULL) {
        ngx_log_error(NGX_LOG_EMERG, cycle->log, 0,
                      "wasmtime: wasm_engine_new() failed");
        return NGX_ERROR;
    }

    wtcf->linker = wasmtime_linker_new(wtcf->engine);
    if (wtcf->linker == NULL) {
        ngx_log_error(NGX_LOG_EMERG, cycle->log, 0,
                      "wasmtime: wasm_linker_new() failed");
        return NGX_ERROR;
    }

    if (wtcf->enable_wasi) {
        error = wasmtime_linker_define_wasi(wtcf->linker);
        if (error != NULL) {
            ngx_wasmtime_log_error(NGX_LOG_EMERG, cycle->log, error, NULL,
                                   "wasmtime: failed to link wasi");
            return NGX_ERROR;
        }
    }

    nhandlers = sizeof(ngx_wasmtime_host_handlers)
                / sizeof(ngx_wasmtime_host_handlers[0]);

    for (i = 0; i < nhandlers; i++) {
        if (ngx_wasmtime_init_host_interface(wtcf, &ngx_wasi_prefix,
                                             &ngx_wasmtime_host_handlers[i],
                                             cycle->log)
            != NGX_OK)
        {
            ngx_log_error(NGX_LOG_EMERG, cycle->log, 0,
                          "wasmtime: failed to init host interface");
            return NGX_ERROR;
        }
    }

    return NGX_OK;
}


static ngx_int_t
ngx_wasmtime_init_host_interface(ngx_wasmtime_conf_t *wtcf, ngx_str_t *ns,
    ngx_wasmtime_host_handler_t *item, ngx_log_t *log)
{
    ngx_uint_t           nargs, i;
    wasm_valtype_t      *arg_types[4], *rtype[1];
    wasm_functype_t     *wft;
    wasmtime_error_t    *error;
    wasm_valtype_vec_t   func_params, results;

    if (item->nargs > 4) {
        return NGX_ERROR;
    }

    nargs = item->nargs;

    for (i = 0; i < nargs; i++) {
        arg_types[i] = wasm_valtype_new_i32();
    }

    wasm_valtype_vec_new(&func_params, nargs, arg_types);

    rtype[0] = wasm_valtype_new_i32();
    wasm_valtype_vec_new(&results, 1, rtype);


    wft = wasm_functype_new(&func_params, &results);

    error = wasmtime_linker_define_func(wtcf->linker,
                                        (char *) ns->data, ns->len,
                                        (char *) item->export.data,
                                        item->export.len,
                                        wft, item->handler, NULL, NULL);
    if (error != NULL) {
        ngx_wasmtime_log_error(NGX_LOG_EMERG, log, error, NULL,
                               "failed to export host function \"%V\"",
                               &item->export);
        return NGX_ERROR;
    }

    return NGX_OK;
}


static ngx_int_t
ngx_wasmtime_enter(ngx_wasm_instance_t *wi, ngx_wasm_call_t *call)
{
    ngx_uint_t                i;
    wasm_trap_t              *trap;
    wasmtime_val_t            res;
    wasmtime_val_t           *fargs;
    wasmtime_func_t           entry;
    wasmtime_error_t         *error;
    ngx_wasmtime_instance_t  *wti;

    if (ngx_wasmtime_lookup_func(wi, &call->name, &entry)
        != NGX_OK)
    {
        ngx_log_error(NGX_LOG_EMERG, wi->log, 0,
                      "ngx_wasmtime_enter failed to lookup function \"%V\"",
                      &call->name);

        return NGX_ERROR;
    }

    wti = wi->data;

    wasmtime_context_set_data(wti->context, call->host);

    ngx_memzero(&res, sizeof(wasmtime_val_t));

    if (call->nargs) {
        fargs = ngx_palloc(call->host->pool,
                           sizeof(wasmtime_val_t) * call->nargs);
        if (fargs == NULL) {
            return NGX_ERROR;
        }

        for (i = 0; i < call->nargs; i++) {

            switch (call->args[i].kind) {
            case NGX_WT_I32:
                fargs[i].kind = WASMTIME_I32;
                fargs[i].of.i32 = call->args[i].of.i32;
                break;

            case NGX_WT_I64:
                fargs[i].kind = WASMTIME_I64;
                fargs[i].of.i64 = call->args[i].of.i64;
                break;

            case NGX_WT_F32:
                fargs[i].kind = WASMTIME_F32;
                fargs[i].of.f32 = call->args[i].of.f32;
                break;

            case NGX_WT_F64:
                fargs[i].kind = WASMTIME_F64;
                fargs[i].of.f64 = call->args[i].of.f64;
                break;

            default:
                return NGX_ERROR;
            }
        }

    } else {
        fargs = NULL;
    }

    trap = NULL;

    error = wasmtime_func_call(wti->context, &entry, fargs, call->nargs,
                               &res, 1, &trap);

    if (error != NULL || trap != NULL) {
        ngx_wasmtime_log_error(NGX_LOG_EMERG, wi->log, error, trap,
                               "failed to call function \"%V\"",
                               &call->name);
        return NGX_ERROR;
    }

    switch (res.kind) {
        case WASMTIME_I32:
            call->rc.of.i32 = res.of.i32;

            ngx_log_debug2(NGX_LOG_DEBUG_CORE, wi->log, 0,
                           "wasmtime: call \"%V\" rc:%D",
                           &call->name, call->rc.of.i32);
            break;

        case WASMTIME_I64:
            call->rc.of.i64 = res.of.i64;

            ngx_log_debug2(NGX_LOG_DEBUG_CORE, wi->log, 0,
                           "wasmtime: call \"%V\" rc:%L",
                           &call->name, call->rc.of.i64);
            break;

        case WASMTIME_F32:
            call->rc.of.f32 = res.of.f32;

            ngx_log_debug2(NGX_LOG_DEBUG_CORE, wi->log, 0,
                           "wasmtime: call \"%V\" rc:%f",
                           &call->name, call->rc.of.f32);
            break;

        case WASMTIME_F64:
            call->rc.of.f64 = res.of.f64;

            ngx_log_debug2(NGX_LOG_DEBUG_CORE, wi->log, 0,
                           "wasmtime: call \"%V\" rc:%f",
                           &call->name, call->rc.of.f64);
            break;

        default:
            ngx_log_error(NGX_LOG_EMERG, wi->log, 0, "unknown wasm type");

            return NGX_ERROR;
    }

    return NGX_OK;
}


static void
ngx_wasmtime_delete_runtime(ngx_cycle_t *cycle)
{
    ngx_wasmtime_conf_t  *wtcf;

    wtcf = (ngx_wasmtime_conf_t *) ngx_wasm_get_conf(cycle->conf_ctx,
                                                     ngx_wasmtime_module);
    wasm_engine_delete(wtcf->engine);
    wtcf->engine = NULL;

    wasmtime_linker_delete(wtcf->linker);
    wtcf->linker = NULL;

    wtcf->config = NULL;
}


static ngx_int_t
ngx_wasmtime_create_module(ngx_cycle_t *cycle, ngx_wasm_module_t *wm)
{
    ngx_wasm_conf_t      *wcf;
    wasmtime_error_t     *error;
    wasmtime_module_t    *mod;
    ngx_wasmtime_conf_t  *wtcf;

    wcf = (ngx_wasm_conf_t *) ngx_wasm_get_conf(cycle->conf_ctx,
                                                ngx_wasm_core_module);

    wtcf = (ngx_wasmtime_conf_t *) ngx_wasm_get_conf(cycle->conf_ctx,
                                                     ngx_wasmtime_module);

    error = wasmtime_module_new(wtcf->engine, (uint8_t *) wm->src.data,
                                wm->src.len, &mod);
    if (error != NULL) {
        ngx_wasmtime_log_error(NGX_LOG_EMERG, cycle->log, error, NULL,
                               "failed to load module");

        ngx_log_error(NGX_LOG_EMERG, cycle->log, 0,
                      "wasmtime: failed to verify module \"%V\"",
                      &wm->file);

        return NGX_ERROR;
    }

    wm->data = mod;
    wm->ctx = wtcf;
    wm->wcf = wcf;

    return NGX_OK;
}


static void
ngx_wasmtime_delete_module(ngx_wasm_module_t *wm)
{
    wasmtime_module_delete(wm->data);
    wm->data = NULL;
}


static ngx_wasm_instance_t *
ngx_wasmtime_create_instance(ngx_wasm_module_t *wm, ngx_pool_t *pool,
    ngx_log_t *log)
{
    wasm_trap_t   *trap;

    wasmtime_error_t   *error;
    wasmtime_extern_t   item;

    ngx_uint_t                i;
    ngx_wasm_preopen_t       *po;
    ngx_wasm_instance_t      *wi;
    ngx_wasmtime_conf_t      *wtcf;
    ngx_wasmtime_instance_t  *wti;

    wtcf = (ngx_wasmtime_conf_t *) wm->ctx;

    wi = ngx_pcalloc(pool, sizeof(ngx_wasm_instance_t));
    if (wi == NULL) {
        return NULL;
    }

    wi->pool = pool;
    wi->log = log;

    wti = ngx_pcalloc(pool, sizeof(ngx_wasmtime_instance_t));
    if (wti == NULL) {
        return NULL;
    }

    wti->store = wasmtime_store_new(wtcf->engine, wtcf, NULL);
    if (wti->store == NULL) {
        ngx_log_error(NGX_LOG_EMERG, wi->log, 0, "wasm_store_new");
        return NULL;
    }

    wti->context = wasmtime_store_context(wti->store);

    wti->wasi_config = wasi_config_new();
    if (wti->wasi_config == NULL) {
        ngx_log_error(NGX_LOG_EMERG, wi->log, 0, "wasi_config");
        return NULL;
    }

    wasi_config_inherit_stdout(wti->wasi_config);
    wasi_config_inherit_stderr(wti->wasi_config);

    if (wm->preopen && wm->preopen->nelts) {

        po = wm->preopen->elts;

        for (i = 0; i < wm->preopen->nelts; i++) {
            if (!wasi_config_preopen_dir(wti->wasi_config,
                                     (char *) po[i].host.data,
                                     (char *) po[i].guest.data))
            {
                ngx_log_error(NGX_LOG_EMERG, wi->log, 0,
                              "failed to preopen host:\"%V\" guest:\"%V\"",
                              &po[i].host, &po[i].guest);
                return NULL;
            }

            ngx_log_debug3(NGX_LOG_DEBUG_CORE, wi->log, 0,
                           "module \"%V\" bind host:\"%V\" to guest:\"%V\"",
                           &wm->ns, &po[i].host, &po[i].guest);
        }
    }

    error = wasmtime_context_set_wasi(wti->context, wti->wasi_config);
    if (error != NULL) {
        ngx_wasmtime_log_error(NGX_LOG_EMERG, wi->log, error, NULL,
                               "failed to instantiate WASI");
        return NULL;
    }

    trap = NULL;

    error = wasmtime_linker_instantiate(wtcf->linker, wti->context, wm->data,
                                        &wti->instance, &trap);

    if (error != NULL || trap != NULL) {
        ngx_wasmtime_log_error(NGX_LOG_EMERG, log, error, trap,
                               "linker failed to create instance");
        return NULL;
    }

    wi->mod = wm;
    wi->data = wti;


    if (!wasmtime_instance_export_get(wti->context, &wti->instance, "memory",
                                      6, &item))
    {
        ngx_log_error(NGX_LOG_EMERG, wi->log, 0,
                      "failed to access instance memory");
        return NULL;
    }

    wti->memory = item.of.memory;

    return wi;
}


static ngx_int_t
ngx_wasmtime_lookup_func(ngx_wasm_instance_t *wi, ngx_str_t *name,
    wasmtime_func_t *func)
{
    ngx_wasmtime_instance_t *wti = wi->data;

    wasmtime_extern_t  item;

    if (!wasmtime_instance_export_get(wti->context, &wti->instance,
                                      (char *) name->data, name->len, &item))
    {
        ngx_log_error(NGX_LOG_EMERG, wi->log, 0,
                      "wasmtime: failed to lookup \"%V\" function", name);
        return NGX_ERROR;
    }

    if (item.kind != WASMTIME_EXTERN_FUNC) {
        ngx_log_error(NGX_LOG_EMERG, wi->log, 0,
                      "wasmtime: \"%V\" is not a function", name);
        return NGX_ERROR;
    }

    *func = item.of.func;

    return NGX_OK;
}


static void
ngx_wasmtime_delete_instance(ngx_wasm_instance_t *wi)
{
    ngx_wasmtime_instance_t  *wti = wi->data;
    wasmtime_store_delete(wti->store);
}


void *
ngx_wasmtime_translate(ngx_wasm_instance_t *wi, uint32_t addr, size_t len)
{
    size_t                    lmem_len;
    uint8_t                  *lmem;
    ngx_wasmtime_instance_t  *wti;

    wti = wi->data;

    lmem = wasmtime_memory_data(wti->context, &wti->memory);
    lmem_len = wasmtime_memory_data_size(wti->context, &wti->memory);

    if (len > lmem_len) {
        return NULL;
    }

    if (addr > (lmem_len - len)) {
        return NULL;
    }

    return &lmem[addr];
}


static wasm_trap_t *
ngx_wasmtime_open(void *env, wasmtime_caller_t *caller,
    const wasmtime_val_t *args, size_t nargs, wasmtime_val_t *rc, size_t nres)
{
    uint8_t          *buf;
    ngx_wasm_host_t  *host;

    host = ngx_wasmtime_call_data(caller);

    buf = ngx_wasm_translate(host->wi, args[0].of.i32, args[1].of.i32);
    if (buf == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->wi->log, 0,
                      "ngx_wasmtime_open bad address");
        rc->of.i32 = NGX_ERROR;
        return NULL;
    }

    rc->of.i32 = host->ops->open(host, buf, args[1].of.i32, args[2].of.i32);

    return NULL;
}


static wasm_trap_t *
ngx_wasmtime_close(void *env, wasmtime_caller_t *caller,
    const wasmtime_val_t *args, size_t nargs, wasmtime_val_t *rc, size_t nres)
{
    ngx_wasm_host_t *host = ngx_wasmtime_call_data(caller);

    rc->of.i32 = host->ops->close(host, args[0].of.i32);

    return NULL;
}


static wasm_trap_t *
ngx_wasmtime_read(void *env, wasmtime_caller_t *caller,
    const wasmtime_val_t *args, size_t nargs, wasmtime_val_t *rc, size_t nres)
{
    uint8_t          *buf;
    ngx_wasm_host_t  *host;

    host = ngx_wasmtime_call_data(caller);

    buf = ngx_wasm_translate(host->wi, args[1].of.i32, args[2].of.i32);
    if (buf == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->wi->log, 0,
                      "ngx_wasmtime_read bad address");
        rc->of.i32 = NGX_ERROR;
        return NULL;
    }

    rc->of.i32 = host->ops->read(host, args[0].of.i32, buf, args[2].of.i32);

    return NULL;
}


static wasm_trap_t *
ngx_wasmtime_write(void *env, wasmtime_caller_t *caller,
    const wasmtime_val_t *args, size_t nargs, wasmtime_val_t *rc, size_t nres)
{
    uint8_t          *buf;
    ngx_wasm_host_t  *host;

    host = ngx_wasmtime_call_data(caller);

    buf = ngx_wasm_translate(host->wi, args[1].of.i32, args[2].of.i32);
    if (buf == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->wi->log, 0,
                      "ngx_wasmtime_write bad address");
        rc->of.i32 = NGX_ERROR;
        return NULL;
    }

    rc->of.i32 = host->ops->write(host, args[0].of.i32, buf, args[2].of.i32);

    return NULL;
}


static wasm_trap_t *
ngx_wasmtime_get(void *env, wasmtime_caller_t *caller,
    const wasmtime_val_t *args, size_t nargs, wasmtime_val_t *rc, size_t nres)
{
    uint8_t          *buf;
    ngx_wasm_host_t  *host;

    host = ngx_wasmtime_call_data(caller);

    buf = ngx_wasm_translate(host->wi, args[2].of.i32, args[3].of.i32);
    if (buf == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->wi->log, 0,
                      "ngx_wasmtime_get bad address");

        rc->of.i32 = NGX_ERROR;
        return NULL;
    }

    rc->of.i32 = host->ops->get(host, args[0].of.i32, args[1].of.i32, buf,
                                args[3].of.i32);

    return NULL;
}


static wasm_trap_t *
ngx_wasmtime_set(void *env, wasmtime_caller_t *caller,
    const wasmtime_val_t *args, size_t nargs, wasmtime_val_t *rc, size_t nres)
{
    uint8_t          *buf;
    ngx_wasm_host_t  *host;

    host = ngx_wasmtime_call_data(caller);

    buf = ngx_wasm_translate(host->wi, args[2].of.i32, args[3].of.i32);
    if (buf == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->wi->log, 0,
                      "ngx_wasmtime_set bad address");
        rc->of.i32 = NGX_ERROR;
        return NULL;
    }

    rc->of.i32 = host->ops->set(host, args[0].of.i32, args[1].of.i32, buf,
                                args[3].of.i32);

    return NULL;
}
